from aiounittest import AsyncTestCase, futurized
from tests.common import example_config
from unittest import mock
from unittest.mock import Mock, patch
from vmshepherd.presets import gitrepo_driver

class MockItem:

    def __init__(self, path):
        self.path = path


class TestGitRepo(AsyncTestCase):

    def setUp(self):
        self.mock_runtime = Mock()
        self.patch_os = patch('vmshepherd.presets.gitrepo_driver.os')
        self.mock_os = self.patch_os.start()

        self.mock_process = Mock()
        self.mock_process.communicate.return_value = futurized(('STDOUT', 'STDERR'))
        self.mock_process.returncode = 0
        self.mock_subprocess_exec = patch(
            'vmshepherd.presets.gitrepo_driver.asyncio.create_subprocess_exec',
            return_value=futurized(self.mock_process)
        ).start()
        self.patch_tmp = patch('vmshepherd.presets.gitrepo_driver.tempfile.TemporaryDirectory')
        self.mock_tmp = self.patch_tmp.start()
        self.mock_tmp.return_value.name = '/tmp/prefixRANDOM_HASH/'

        self.config = {
            'repositories': {
                'paas': 'git://testrepm/paas.git',
                'db': 'git://stash/db.git',
            },
            'clone_dir': '/tmp/'
        }
        self.driver = gitrepo_driver.GitRepoDriver(self.config, self.mock_runtime, example_config['defaults'])

    def tearDown(self):
        patch.stopall()

    def test_init(self):
        self.assertEqual('/tmp/', self.driver._clone_dir)
        self.assertEqual(None, self.driver._presets)

    async def test_clone(self):
        self.mock_os.path.exists.return_value = False

        await self.driver._clone_or_update('/tmp', 'blah')

        self.mock_process.communicate.assert_called_once_with()
        self.mock_subprocess_exec.assert_called_once_with('git', 'clone', 'blah', '/tmp', stderr=-1, stdout=-1)

    async def test_update(self):
        self.mock_os.path.exists.return_value = True

        await self.driver._clone_or_update('/tmp', 'blah')

        self.mock_process.communicate.assert_called_once_with()
        self.mock_subprocess_exec.assert_called_once_with('git', '-C', '/tmp', 'pull', stderr=-1, stdout=-1)

    async def test_update_fail(self):
        self.mock_os.path.exists.return_value = True
        self.mock_process.returncode = 1

        with self.assertRaises(RuntimeError):
            await self.driver._clone_or_update('/tmp', 'blah')

        self.mock_process.communicate.assert_called_once_with()
        self.mock_subprocess_exec.assert_called_once_with('git', '-C', '/tmp', 'pull', stderr=-1, stdout=-1)

    async def test_load_repos_presets(self):
        self.patch_os.stop()
        res = await self.driver._load_repo('blah', example_config['presets']['path'])
        self.assertEqual(res['blah.test-preset']['count'], 1)
        self.assertEqual(res['blah.test-preset']['name'], 'blah.test-preset')
        self.assertEqual(res['blah.test-preset']['flavor'], 'm1.small')
        self.assertEqual(res['blah.test-preset']['image'], 'fedora-27')
        self.assertEqual(res['blah.test-preset']['meta_tags'], {'key1': 'value1'})

    async def test_reload(self):
        self.mock_os.path.join.side_effect = lambda a, b: (a + b)
        self.driver._clone_or_update = Mock(return_value=futurized(None))
        self.driver._assure_clone_dir_exists = Mock(return_value=True)
        self.driver._load_repo = Mock(return_value=futurized({}))
        await self.driver.list_presets()
        self.driver._load_repo.assert_has_calls([
            mock.call('paas', '/tmp/paas'), mock.call('db', '/tmp/db')
        ])
        self.assertEqual(self.driver._load_repo.call_count, 2)

    async def test_reload_tmp_path(self):
        self.mock_os.path.exists.return_value = False
        cfg = {
            'repositories': {
                'paas': 'git://testrepm/paas.git',
                'db': 'git://stash/db.git',
            }}
        driver = gitrepo_driver.GitRepoDriver(cfg, self.mock_runtime, example_config['defaults'])
        self.mock_os.path.join.side_effect = lambda a, b: (a + b)
        driver._clone_or_update = Mock(return_value=futurized(None))
        driver._assure_clone_dir_exists = Mock(return_value=True)
        driver._load_repo = Mock(return_value=futurized({}))
        await driver.list_presets()
        driver._load_repo.assert_has_calls([
            mock.call('paas', '/tmp/prefixRANDOM_HASH/paas'), mock.call('db', '/tmp/prefixRANDOM_HASH/db')
        ])
        self.assertEqual(driver._load_repo.call_count, 2)



import os
from .abstract import AbstractConfigurationDriver
from vmshepherd.utils import async_load_from_yaml_file


class DirectoryDriver(AbstractConfigurationDriver):

    def __init__(self, config, runtime, defaults):
        super().__init__(runtime, defaults)
        self._presets = {}
        self._path = config['path']

    async def _get_preset_spec(self, preset_name: str):
        return self._specs[preset_name]

    async def _list(self):
        await self._reload()
        return self._specs.keys()

    async def _reload(self):
        _tmp_specs = {}
        for item in os.scandir(self._path):
            if os.path.isfile(item.path) and item.path.endswith('.conf'):
                config = await async_load_from_yaml_file(item.path)
                config['name'] = preset_name = f"{config['name']}"
                config['userdata_source_root'] = self._path
                _tmp_specs[preset_name] = config
        self._specs = _tmp_specs

    def reconfigure(self, config, defaults):
        super().reconfigure(config, defaults)
        self._path = config.get('path', self._path)



import aiohttp_jinja2
import jinja2
import logging
import os
import time
from .rpc_api import RpcApi
from aiohttp import web


class WebServer(web.Application):

    def __init__(self, vmshepherd, config=None):
        super().__init__()
        self.port = config.get('listen_port', 8888)
        allowed_methods = config.get('api', {}).get('allowed_methods')
        self.vmshepherd = vmshepherd
        self.configure_panel()
        self.configure_api(allowed_methods)

    def configure_panel(self):
        """
        Configure templates and routing
        """
        webroot = os.path.dirname(__file__)

        self.template_path = os.path.join(webroot, 'templates')
        aiohttp_jinja2.setup(
            self, loader=jinja2.FileSystemLoader(self.template_path),
            filters={'sorted': sorted, 'int': int}
        )

        self['static_root_url'] = '/static'
        self.router.add_view('/', Panel)
        self.router.add_static(
            '/static/', path=os.path.join(webroot, 'static'), name='static'
        )

    def configure_api(self, allowed_methods=None):
        """
        Configure route for api
        :arg list allowed_methods: List of a methods which are turned on in our api
        """
        self.router.add_route('POST', '/api', RpcApi(allowed_methods).handler)

    async def start(self):
        """
        Initialize and start WebServer
        """
        logging.info('Starting server, listening on %s.', self.port)
        runner = web.AppRunner(self)
        await runner.setup()
        site = web.TCPSite(runner, '', self.port)
        await site.start()


class Panel(web.View):

    @aiohttp_jinja2.template('index.html.jinja2')
    async def get(self):
        """
        Inject all preset data to Panel and Render a Home Page
        """
        shepherd = self.request.app.vmshepherd
        data = {'presets': {}, 'config': shepherd.config}
        presets = await shepherd.preset_manager.list_presets()
        runtime = shepherd.runtime_manager
        for name in presets:
            preset = shepherd.preset_manager.get_preset(name)
            await preset.refresh_vms(if_older_than=60)
            data['presets'][name] = {
                'preset': preset,
                'vms': preset.vms,
                'runtime': await runtime.get_preset_data(name),
                'vmshepherd_id': shepherd.instance_id,
                'now': time.time()
            }
        return data



import asyncio
from aiohttp_jsonrpc.client import ServerProxy, batch

# initialize a loop
loop = asyncio.get_event_loop()
client = ServerProxy("http://127.0.0.1:8888/api", loop=loop)

# lets list a virtual machines
async def main():
    print(await client.list_vms(preset='C_DEV-app-dev'))

# close a client
client.close()
loop.run_until_complete(main())



import asyncio
import logging
import time
from collections import Counter
from vmshepherd.iaas.vm import VmState


class Preset:

    def __init__(self, name: str):
        self.iaas = None
        self.healthcheck = None
        self.unmanaged = False
        self.runtime_mgr = None
        self.runtime = None
        self.config = {}
        self.name = name
        self.count = 0
        self.created = 0
        self.terminated = 0
        self.healthcheck_terminated = 0
        self._extra = {'preset': self.name}
        self._locked = False
        self._vms = []
        self._vms_refresh_time = 0

    @property
    def vms(self):
        return self._vms

    def configure(self, config: dict, runtime_mgr: object, iaas: object, healthcheck: object):
        self.iaas = iaas
        self.healthcheck = healthcheck
        self.unmanaged = config.get('unmanaged', False)
        self.config = config
        self.count = config['count']
        self.runtime_mgr = runtime_mgr

    def _reset_counters(self):
        self.created = 0
        self.terminated = 0
        self.healthcheck_terminated = 0

    async def _terminate_vm(self, vm):
        await vm.terminate()

    async def __aenter__(self):
        self.runtime = await self.runtime_mgr.get_preset_data(self.name)
        require_manage = time.time() - self.runtime.last_managed > self.config.get('manage_interval', 60)
        if not require_manage:
            return False

        expired = time.time() - self.runtime.last_managed > self.config.get('manage_expire', 120)
        self._locked = expired or (await self.runtime_mgr.acquire_lock(self.name))

        self._reset_counters()

        return self._locked

    async def __aexit__(self, exc_type, exc, tb):
        if self._locked:
            self._locked = False
            await self.runtime_mgr.set_preset_data(self.name, self.runtime)
            await self.runtime_mgr.release_lock(self.name)

    async def _create_vms(self, count):
        for i in range(count):
            try:
                args = dict(
                    preset_name=self.name, image=self.config['image'], flavor=self.config['flavor'],
                    userdata=self.config.get('userdata'), key_name=self.config.get('key_name'),
                    **(self.config.get('network', {}))
                )
                await self.iaas.create_vm(**args)
                self.created += 1
            except Exception:
                logging.error('Could not create vm with %s', args, extra=self._extra)

    async def refresh_vms(self, if_older_than=None):
        """ Refresh vm list

        :arg integer if_older_than: Interval in seconds you accept cached data
        """
        if if_older_than is None or time.time() - self._vms_refresh_time > if_older_than:
            self._vms_refresh_time = time.time()
            self._vms = await self.iaas.list_vms(self.name)

    async def manage(self):
        """ Manage function docstring"""

        await self.refresh_vms()

        vms_stat = Counter([vm.get_state() for vm in self._vms])
        missing = self.count - len(self._vms) if len(self._vms) < self.count else 0
        logging.info(
            'State: %s, VMs Status: %s expected, %s in iaas, %s running, %s nearby shutdown, %s pending, '
            '%s after time shutdown, %s terminated, %s error, %s unknown, %s missing',
            'unmanaged' if self.unmanaged else 'managed', self.count, len(self._vms), vms_stat[VmState.RUNNING.value],
            vms_stat[VmState.NEARBY_SHUTDOWN.value], vms_stat[VmState.PENDING.value],
            vms_stat[VmState.AFTER_TIME_SHUTDOWN.value], vms_stat[VmState.TERMINATED.value],
            vms_stat[VmState.ERROR.value], vms_stat[VmState.UNKNOWN.value], missing, extra=self._extra
        )

        to_create = 0
        if not self.unmanaged:
            for vm in self._vms:
                if vm.is_dead():
                    logging.info("Terminate %s", vm, extra=self._extra)
                    await vm.terminate()
                    self.terminated += 1
            to_create = self.count - (len(self._vms) - self.terminated - vms_stat[VmState.NEARBY_SHUTDOWN.value])
            logging.debug("Create %s Vm", to_create, extra=self._extra)
            await self._create_vms(to_create)

        await self._healthcheck(self._vms)

        if not self.unmanaged:
            logging.info(
                'VMs Status update: %s terminated, %s terminated by healthcheck, %s created, %s failed healthcheck',
                self.terminated, self.healthcheck_terminated, to_create, len(self.runtime.failed_checks),
                extra=self._extra
            )
        else:
            logging.info(
                'VMs Status update: %s failed healthcheck', len(self.runtime.failed_checks), extra=self._extra
            )

    async def _healthcheck(self, vms):
        _healthchecks = {}
        for vm in vms:
            if not vm.is_dead():
                _healthchecks[vm] = asyncio.ensure_future(self.healthcheck.is_healthy(vm))
        await asyncio.gather(*list(_healthchecks.values()), return_exceptions=True)
        current_fails = []

        for vm, state_check in _healthchecks.items():
            # if check failed
            if not state_check.result():
                current_fails.append(vm.id)
                failed_since = self.runtime.failed_checks.get(vm.id, {}).get('time', time.time())
                count_fails = self.runtime.failed_checks.get(vm.id, {}).get('count', 0)
                self.runtime.failed_checks[vm.id] = {'time': failed_since, 'count': count_fails + 1}
                terminate_heatlh_failed_delay = self.config.get('healthcheck', {}).get('terminate_heatlh_failed_delay', -1)
                if not self.unmanaged and terminate_heatlh_failed_delay >= 0 and count_fails > 5:
                    if terminate_heatlh_failed_delay + failed_since < time.time():
                        logging.debug("Terminate %s, healthcheck fails (count %s) since %s", vm, count_fails,
                                      failed_since, extra=self._extra)
                        await self._terminate_vm(vm)
                        self.healthcheck_terminated += 1

        for vm_id in list(self.runtime.failed_checks.keys()):
            if vm_id not in current_fails:
                del self.runtime.failed_checks[vm_id]



''' Preset configuration (or sometime called Preset store) is a mechanism to fetch and prepare cluster definition/spec for VmShepherd.
Preset configuration driver should implement `_get_preset_spec` and `_list`.

Initialization - consider following config:

::

   presets:
     driver: SomeC
     param1: AAAA
     param2: BBBB
     some_x: CCC

All params will be passed as config dict to the driver init:

'''
import os
from .preset import Preset
from jinja2 import Template
from vmshepherd.drivers import Drivers
from vmshepherd.utils import get_merged_dict_recursively, async_load_from_file
from vmshepherd.errors import PresetNotFound

class AbstractConfigurationDriver:

    def __init__(self, runtime, defaults):
        self.runtime = runtime
        self.defaults = defaults
        self._presets = None

    async def _get_preset_spec(self, preset_name: str):
        """ Returns configuration of specific preset

        :arg string preset_name: Name of the preset

        Returns dict - preset configuration

        Example configuration:

        ::

            name: 'C_DEV-app-dev'
            count: 1
            flavor: m1.small
            image: ubuntu-xenial
            manage_interval: 0    # interval in seconds
            manage_expire: 120    # max time of manage (after this time another manage can be scheduled) - expiration in seconds
            network:
              security_groups:
                - PaaS PrivateCloud
                - paas
              availability_zone: nova
              subnet:
                - pl-krk-2-something
            iaas:
              project_name: PROJECT_NAME
              project_domain_name: PROJECT_DOMAIN
            userdata: ''
            meta_tags: {}

        """
        raise NotImplementedError

    async def _list(self):
        """ Returns list names of presets to manage

        Returns list of string - just names. Further information will be collected with given name and `_get_preset_spec` method.
        """
        raise NotImplementedError

    def get_preset(self, preset_name):
        try:
            return self._presets[preset_name]
        except KeyError:
            raise PresetNotFound(preset_name)

    async def list_presets(self, refresh=True):
        if refresh or self._presets is None:
            await self.refresh_presets()
        return self._presets

    async def refresh_presets(self):
        if self._presets is None:
            self._presets = {}

        fresh_presets = set(await self._list())
        loaded_presets = set(self._presets.keys())
        to_add = fresh_presets - loaded_presets
        to_remove = loaded_presets - fresh_presets

        for name in to_add:
            self._presets[name] = Preset(name)
        for name in to_remove:
            del self._presets[name]

        for name, preset in self._presets.items():
            spec = await self._get_preset_spec(name)
            params = await self._prepare_preset_params(preset, spec)
            preset.configure(**params)

    async def _prepare_preset_params(self, preset, config):
        iaas_cfg = get_merged_dict_recursively(
            self.defaults.get('iaas', {}), config.get('iaas', {})
        )
        iaas = Drivers.get('iaas', iaas_cfg)
        config['iaas'] = iaas_cfg

        healthcheck_cfg = get_merged_dict_recursively(
            self.defaults.get('healthcheck', {}), config.get('healthcheck', {})
        )
        healthcheck = Drivers.get('healthcheck', healthcheck_cfg)
        config['healthcheck'] = healthcheck_cfg

        if 'userdata' in config:
            await self._inject_preset_userdata(config)
            self._render_preset_userdata(config)

        return {
            'config': config, 'runtime_mgr': self.runtime,
            'iaas': iaas, 'healthcheck': healthcheck
        }

    def reconfigure(self, config, defaults):
        self.defaults = defaults

    async def _inject_preset_userdata(self, config):
        if 'userdata' not in config or not config['userdata']:
            return
        root = config.get('userdata_source_root', '')

        if config['userdata'].startswith('file://'):
            path = os.path.join(root, config['userdata'].replace('file://', ''))
            config['userdata'] = await async_load_from_file(path)

    def _render_preset_userdata(self, config):
        tpl = Template(config['userdata'])
        config['userdata'] = tpl.render(**config)



import base64
import logging
import time
from .abstract import AbstractIaasDriver
from vmshepherd.errors import IaaSError
from .vm import Vm, VmState
from asyncopenstackclient import NovaClient, GlanceClient, AuthPassword
from bidict import bidict
from datetime import datetime
from simplejson.errors import JSONDecodeError


class OpenStackDriver(AbstractIaasDriver):

    def __init__(self, config):
        self.config = config

    def openstack_exception(func):
        '''
            Openstack exceptions decorator
        '''
        async def wrap(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                logging.error(e)
                raise IaaSError(memoryview='OpenStackError')
        return wrap

    def initialize_openstack(func):
        '''
            Initialize and refresh openstack connection

        '''
        async def wrap(self, *args, **kwargs):
            if not hasattr(self, 'auth') or not self.auth.is_token_valid():
                self.auth = AuthPassword(auth_url=self.config['auth_url'],
                                         username=self.config['username'],
                                         password=self.config['password'],
                                         project_name=self.config['project_name'],
                                         user_domain_name=self.config['user_domain_name'],
                                         project_domain_name=self.config['project_domain_name'])
                self.nova = NovaClient(session=self.auth)
                self.glance = GlanceClient(session=self.auth)
                await self.nova.init_api(timeout=self.config.get('http_timeout', 10))
                await self.glance.init_api(timeout=self.config.get('http_timeout', 10))

            if not hasattr(self, 'last_init') or self.last_init < (time.time() - 60):
                await self.initialize()
                self.last_init = time.time()
            return await func(self, *args, **kwargs)
        return wrap

    async def initialize(self):
        '''
         Initialize static data like images and flavores and set it as object property
        '''
        flavors = await self._list_flavors()
        images = await self._list_images()

        self.flavors_map = bidict()
        self.images_map = bidict()
        self.images_details = {}

        for flavor in flavors:
            self.flavors_map.put(flavor['id'], flavor['name'], on_dup_key='OVERWRITE', on_dup_val='OVERWRITE')

        for image in images:
            # @TODO filetes :
            # @TODO filtering by owner
            # if hasattr(image, 'owner_id') and  image.owner_id in self.config['image_owner_ids']:
            #  @TODO enable filtering by tag
            # if 'lastest' in image.tags:
            self.images_details[image['id']] = {
                'name': image['name'],
                'created_at': image['created_at'],
                'latest': 'latest' in image['tags']
            }
            self.images_map.put(image['id'], image['name'], on_dup_key='OVERWRITE', on_dup_val='OVERWRITE')

    @initialize_openstack
    @openstack_exception
    async def create_vm(self, preset_name, image, flavor, security_groups=None,
                        userdata=None, key_name=None, availability_zone=None, subnets=None):
        '''
          Create VM
         :arg preset_name: string
         :arg image: string image id
         :arg flavor: string flavor id
         :arg security_groups:  list
         :arg userdata: string
         :arg key_name: string
         :arg availability_zone: string
         :arg subnets: list
         :returns list Vm objects
         @TODO
         1. returns image id
        '''
        image_id = self.images_map.inv.get(image)
        flavor_id = self.flavors_map.inv.get(flavor)
        spec = {
            "name": preset_name,
            "flavorRef": flavor_id,
            "imageRef": image_id,
            "security_groups": [{"name": group} for group in security_groups],
            "user_data": userdata
        }
        if availability_zone is not None:
            spec.update({"availability_zone": availability_zone})
        if subnets is not None:
            spec.update({"networks": [{'uuid': subnet['net-id']} for subnet in subnets]})
        if userdata is not None:
            userdata = userdata.encode('utf-8')
            userdata = base64.b64encode(userdata).decode('utf-8')
            spec.update({"user_data": userdata})

        result = await self.nova.servers.create(server=spec)
        return result["server"]

    @initialize_openstack
    @openstack_exception
    async def list_vms(self, preset_name):
        '''
        List VMs by preset name
        :arg present_name: string
        '''

        response = await self.nova.servers.list(name=f'^{preset_name}$')
        result = []
        for server in response['servers']:
            result.append(self._map_vm_structure(server))
        return result

    @openstack_exception
    async def terminate_vm(self, vm_id):
        '''
         Terminate VM
         :arg vm_id: string
        '''
        try:
            await self.nova.servers.force_delete(vm_id)
        except JSONDecodeError as exc:
            logging.info("nova sent 'content-type: application/json' but no content appeared, whatever")
            pass
        except Exception:
            raise

    @initialize_openstack
    @openstack_exception
    async def get_vm(self, vm_id):
        '''
        Get VM
        :arg vm_id: string
        :returns vm: object
        '''
        result = await self.nova.servers.get(vm_id)
        return self._map_vm_structure(result["server"])

    @openstack_exception
    async def _list_flavors(self):
        '''
        Returns list of flavors from Openstack
        '''
        result = await self.nova.flavors.list()
        return result['flavors']

    @openstack_exception
    async def _list_images(self):
        '''
        Returns list of images from OpenStack
        '''
        result = await self.glance.images.list()
        return result['images']

    def _map_vm_structure(self, vm):
        '''
        Vm unification
        :arg vm: object
        :returns object
        '''
        ip = self._extract_ips(vm['addresses'])
        created = datetime.strptime(vm['created'], '%Y-%m-%dT%H:%M:%SZ')
        flavor = self.flavors_map.get(vm['flavor'].get('id'))
        image = self.images_map.get(vm['image'].get('id'))
        timed_shutdown_at = vm.get('metadata', {}).get('iaas_timed_shutdown')
        timed_shutdown_at = int(timed_shutdown_at) if timed_shutdown_at else None
        state = self._map_vm_status(vm['status'], timed_shutdown_at)
        iaasvm = Vm(self, vm['id'], vm['name'], ip, created, state=state, metadata=vm['metadata'], tags=vm.get('tags', []), flavor=flavor,
                    image=image, timed_shutdown_at=timed_shutdown_at)
        return iaasvm

    def _map_vm_status(self, openstack_status, timed_shutdown_at=None):
        '''
         Map openstack vm statuses to vmshepherd vm statuses
         openstack vm statuses: ACTIVE, BUILD, DELETED, ERROR, HARD_REBOOT, MIGRATING, PASSWORD, PAUSED, REBOOT,
         REBUILD, RESCUE, RESIZE, REVERT_RESIZE, SHELVED, SHELVED_OFFLOADED, SHUTOFF, SOFT_DELETED, SUSPENDED, UNKNOWN,
         VERIFY_RESIZE

         :arg string openstack_status
         :returns string
        '''
        statuses = {
            VmState.TERMINATED: [
                'ERROR',
                'DELETED',
                'SHUTOFF',
                'SOFT_DELETED',
                'SUSPENDED'
            ],
            VmState.PENDING: [
                'BUILD',
                'REBUILD'
            ],
            VmState.RUNNING: ['ACTIVE']
        }

        state = VmState.UNKNOWN
        for vmstate, value in statuses.items():
            if openstack_status in value:
                state = vmstate
                break
        if timed_shutdown_at:
            now = time.time()
            if timed_shutdown_at < now:
                state = VmState.AFTER_TIME_SHUTDOWN
            elif (timed_shutdown_at - now) < self.config.get('shutdown_grace_period', 900):
                state = VmState.NEARBY_SHUTDOWN

        return state

    def _extract_ips(self, data):
        '''
        Extract ip addressess from openstack structure
        {
          'pl-krk-2-int-301-c2-int-1': [
           {
             'OS-EXT-IPS-MAC:mac_addr': 'fa:16:3e:29:f1:bb',
             'version': 4,
             'addr': '10.185.138.36',
             'OS-EXT-IPS:type': 'fixed'
          }
         ]
       }
       :arg data: dict
       :returns list
       '''
        result = []
        for region in data.items():
            for interface in region[1]:
                result.append(interface['addr'])
        return result



import aiofiles
import asyncio
import collections
import copy
import logging
import random
import socket
import string
import yaml


def get_merged_dict_recursively(d, u):
    new = copy.deepcopy(d)
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            new[k] = get_merged_dict_recursively(new.get(k, {}), v)
        else:
            new[k] = v
    return new


def load_config_file(path):
    with open(path, 'r') as f:
        config = yaml.safe_load(f)
    return config


async def async_load_from_file(path):
    async with aiofiles.open(path, mode='r') as f:
        data = await f.read()
    return data


async def async_load_from_yaml_file(path):
    contents = await async_load_from_file(path)
    return yaml.safe_load(contents)


def add_async_delay(func):
    async def wrapper(*args, **kwargs):
        await asyncio.sleep(0.1)
        return await func(*args, **kwargs)
    return wrapper


def next_id():
    i = 0
    while True:
        yield i
        i += 1


def gen_id(rnd_length):
    hostname = socket.gethostname()
    rnd = ''.join(random.choice(
        string.ascii_lowercase + string.digits) for i in range(rnd_length)
    )
    return f'{hostname}-{rnd}'


class PrefixFilter(logging.Filter):
    """Prepends prefix to logged message."""
    def __init__(self, prefix):
        self._prefix = prefix

    def filter(self, record):
        if hasattr(record, 'preset'):
            record.msg = '[%s][%s][%s]: %s' % (self._prefix, record.preset, record.module, record.msg)
        else:
            record.msg = '[%s][%s]: %s' % (self._prefix, record.module, record.msg)
        return True


def prefix_logging(prefix, handler=None):
    if handler is not None:
        handlers = [handler]
    else:
        handlers = logging.getLogger().handlers
    for hndlr in handlers:
        hndlr.addFilter(PrefixFilter(prefix))



import aiohttp
import logging
from .abstract import AbstractHealthcheck


class HttpHealthcheck(AbstractHealthcheck):

    def __init__(self, config=None):
        config = config or {}
        self.method = config.get('method', 'GET')
        self.path = config.get('path', '/')
        self.port = config.get('port', 80)
        self.check_status = config.get('check_status', 200)
        self._client_timeout = aiohttp.ClientTimeout(
            connect=float(config.get('conn_timeout', 1)),
            total=float(config.get('read_timeout', 1))
        )

    def __str__(self):
        return f'HTTP on {self.port} with {self.method} {self.path} expecting {self.check_status}'

    async def is_healthy(self, vm):
        if not vm.ip:
            logging.info('[healthcheck] ip is empty:%s', vm)
            return False

        check_status = False
        try:
            async with aiohttp.ClientSession(timeout=self._client_timeout) as session:
                ip = vm.ip[0] if type(vm.ip) == list else vm.ip
                url = "http://{}:{}{}".format(ip, self.port, self.path)
                resp = await session.request(method=self.method, url=url)
                check_status = resp.status == self.check_status
        except aiohttp.client_exceptions.ClientError:
            pass
        logging.debug('[healthcheck] VM: %s - check status: %s', vm, check_status)
        return check_status



import asyncio
import logging
import os
from vmshepherd.drivers import Drivers
from vmshepherd.http import WebServer
from vmshepherd.utils import gen_id, prefix_logging
from vmshepherd.worker import Worker


class VmShepherd:

    def __init__(self, config):
        self.config = config
        self.root_dir = os.path.dirname(__file__)
        self.instance_id = gen_id(rnd_length=5)
        self.setup_logging()

        self.runtime_manager = Drivers.get(
            'runtime', self.config['runtime'],
            instance_id=self.instance_id
        )

        self.preset_manager = Drivers.get(
            'presets', self.config['presets'],
            runtime=self.runtime_manager,
            defaults=self.config.get('defaults', {})
        )

        self.worker = Worker(
            runtime=self.runtime_manager, presets=self.preset_manager,
            interval=int(self.config.get('worker_interval', 5)),
            autostart=self.config.get('autostart', True)
        )

        http_conf = self.config.get('http', None)
        if http_conf:
            self.web = WebServer(self, http_conf)
            asyncio.ensure_future(self.web.start())

    async def run(self, run_once=False):
        if run_once:
            await self.worker.run_once()
        else:
            await self.worker.run_forever()

    def setup_logging(self):
        logger = logging.getLogger()
        log_level = self.config.get('log_level', 'info').upper()
        logger.setLevel(log_level)
        if logger.getEffectiveLevel() == logging.DEBUG:
            logging.debug('DEBUG mode enabled')
        prefix_logging(self.instance_id)

    def reload(self, with_config=None):
        self.config = with_config or self.config
        self.runtime_manager.reconfigure(self.config.get('runtime'))
        self.preset_manager.reconfigure(self.config.get('presets'), self.config.get('defaults'))
        Drivers.flush()



import os.path

root_dir = os.path.dirname(__file__)


example_config = {
    'autostart': False,
    'runtime': {
        'driver': 'InMemoryDriver',
    },
    'presets': {
        'driver': 'DirectoryDriver',
        'path': os.path.join(root_dir, 'presets.d')
    },
    'defaults': {
        'iaas': {
            'driver': 'DummyIaasDriver',
            'api_key': 'asd',
            'api_secret': 'asd'
        },
        'healthcheck': {
            'driver': 'HttpHealthcheck',
            'conn_timeout': 1,
            'read_timeout': 1,
            'port': 7007,
            'path': '/',
            'method': 'GET',
            'check_status': 200,
            'terminate_heatlh_failed_delay': 1000
        }
    }
}



from aiounittest import futurized, AsyncTestCase
from bidict import bidict
from unittest.mock import patch
from vmshepherd.iaas import OpenStackDriver


class MockVM:
    def __init__(self):
        pass


class TestOpenStackDriver(AsyncTestCase):

    def setUp(self):
        self.config = {
            'auth_url': 'http://keystone.api.int.iaas:5000/v3',
            'username': 'paas-c2-int-1-api-access',
            'password': '00e91e6550bcd29cfd306280df1caf5e',
            'project_id': '7b6052ae16434e79bc6456808c9f37a6',
            'project_name': 'c2-int-1',
            'api_version': '2.26',
            'flavor_id': '9c17893c-6fc4-4a69-96c8-6e5a8af6b836',
            'image_id': 'c0d3fe3e-52f9-4a2c-aeb1-b8c977fd7d17',
            'flavor': 'm1.tiny',
            'image': 'ubuntu-xenial',
            'project_domain_name': 'grupa.onet',
            'user_domain_name': 'default',
            'image_owner_ids': [
                'bb321996cce74e549c4ae00a67b657b8'
            ]
        }

        self.vm = {
            'name': 'test-vm-name',
            'id': '099fds8f9ds89fdsf',
            'metadata': {},
            'status': 'ACTIVE',
            'created': '2018-02-02T14:29:49Z',
            'updated': '2018-02-02T14:30:00Z',
            'tags': [],
            'addresses': {
                'pl-krk-2-int-301-c2-int-1': [
                    {
                        'OS-EXT-IPS-MAC:mac_addr': 'fa:16:3e:29:f1:bb',
                        'version': 4,
                        'addr': '10.185.138.36',
                        'OS-EXT-IPS:type': 'fixed'
                    }
                ]
            },
            'flavor': {'id': 'testflavorid'},
            'image': {'id': 'testimgid'}
        }

    async def test_map_vm_structure(self):
        '''
        Test if mapped structure is correct
        '''
        mock_list_flavors = patch('vmshepherd.iaas.OpenStackDriver._list_flavors').start()
        mock_list_flavors.return_value = futurized([])
        mock_list_images = patch('vmshepherd.iaas.OpenStackDriver._list_images').start()
        mock_list_images.return_value = futurized([])

        osd = OpenStackDriver(self.config)

        osd.flavors_map = bidict()
        osd.flavors_map['testflavorid'] = 'testflavorname'
        osd.images_map = bidict()
        osd.images_map['testimgid'] = 'testimgname'

        result = osd._map_vm_structure(self.vm)

        self.assertEqual(result.id, '099fds8f9ds89fdsf')
        self.assertEqual(result.name, 'test-vm-name')
        self.assertEqual(result.metadata, {})
        self.assertEqual(str(result.state), 'VmState.RUNNING')
        self.assertEqual(result.ip, ['10.185.138.36'])
        self.assertEqual(result.flavor, 'testflavorname')
        self.assertEqual(result.image, 'testimgname')

    async def test_initialize(self):
        '''
        test if initialize method set flavors and images properties
        '''
        image = {
            'id': 'imageid',
            'name': 'imagename',
            'created_at': '2018-02-02T14:30:00Z',
            'tags': ['latest']
        }
        flavor = {
            'id': 'flavorid',
            'name': 'flavorname'
        }
        mock_list_flavors = patch('vmshepherd.iaas.OpenStackDriver._list_flavors').start()
        mock_list_flavors.return_value = futurized([flavor])
        mock_list_images = patch('vmshepherd.iaas.OpenStackDriver._list_images').start()
        mock_list_images.return_value = futurized([image])

        osd = OpenStackDriver(self.config)

        await osd.initialize()
        self.assertEqual(osd.flavors_map['flavorid'], 'flavorname')
        self.assertEqual(osd.images_map['imageid'], 'imagename')



class VmShepherdError(Exception):
    message = None
    details = None

    def __init__(self, message, details=''):
        super().__init__(self.message)
        self.message = message
        self.details = details
        assert self.message is not None

    def __str__(self):
        if self.details:
            return f'VmShepherdError: {self.message}, details={self.details}'
        return f'VmShepherdError: {self.message}'

    def __repr__(self):
        if self.details:
            return f'VmShepherdError(message={self.message}, details={self.details}'
        return f'VmShepherdError(message={self.message})'


class PresetNotFound(VmShepherdError):
    def __init__(self, preset_name):
        """
        :param preset_name: Preset human friendly id
        """
        self.message = f"Preset {preset_name} not found"
        super().__init__(self.message)


class IaaSError(VmShepherdError):
    def __init__(self, message='IaaSError', details=None):
        super().__init__(message, details)


class IaasPresetConfigurationError(IaaSError):
    def __init__(self, message='IaaSPresetConfigurationError', details=None):
        super().__init__(message, details)


class IaasCommunicationError(IaaSError):
    def __init__(self, message='IaaSPresetConfigurationError', details=None):
        super().__init__(message, details)


class IaasAuthError(IaaSError):
    def __init__(self, message='IaaSAuthError', details=None):
        super().__init__(message, details)


class DummyIaasUserError(IaaSError):
    def __init__(self, message="DummyIaasUserError", details=None):
        super().__init__(message, details)


class DummyIaasVmNotFound(DummyIaasUserError):
    def __init__(self):
        super().__init__('VMNOTFOUND', 'Vm Not Found')



import asyncio
from .common import example_config
from aiounittest import AsyncTestCase, futurized
from collections import namedtuple
from unittest import expectedFailure
from unittest.mock import Mock
from vmshepherd import VmShepherd
from vmshepherd.drivers import Drivers
from vmshepherd.iaas import VmState

VmMock = namedtuple('VmMock', 'id ip name image flavor')


class TestVmShepherdInitWithDummyDrivers(AsyncTestCase):

    def setUp(self):
        super().setUp()
        Drivers.flush()
        self.vmshepherd = VmShepherd(example_config)

    async def test_initialize_preset_manager(self):
        presets_list = await self.vmshepherd.preset_manager.list_presets()
        self.assertEqual(list(presets_list.keys()), ['test-preset'])

        preset = self.vmshepherd.preset_manager.get_preset('test-preset')
        self.assertEqual(preset.count, 1)

    async def test_initialize_iaas(self):
        await self.vmshepherd.runtime_manager.acquire_lock('test-preset')
        await self.vmshepherd.runtime_manager.release_lock('test-preset')


class TestVmShepherdRunWithDummyDrivers(AsyncTestCase):

    def setUp(self):
        super().setUp()
        Drivers.flush()
        self.vmshepherd = VmShepherd(example_config)

    async def test_run(self):
        # zero run - should create on vm
        await self.vmshepherd.run(run_once=True)

        # first run
        await self.vmshepherd.run(run_once=True)

        preset = self.vmshepherd.preset_manager.get_preset('test-preset')
        self.assertEqual(
            preset.vms,
            [VmMock(0, ['127.0.0.1'], 'test-preset', 'fedora-27', 'm1.small')]
        )
        self.assertTrue(preset.vms[0].is_running())

        # # second run - there should be no change
        await self.vmshepherd.run(run_once=True)

        preset = self.vmshepherd.preset_manager.get_preset('test-preset')
        self.assertEqual(
            preset.vms,
            [VmMock(0, ['127.0.0.1'], 'test-preset', 'fedora-27', 'm1.small')]
        )
        self.assertTrue(preset.vms[0].is_running())

        # third run - virtual machine goes in ERROR
        #  - failed vm should be terminated
        #  - new vm should schdule new vm

        preset.vms[0].state = VmState.ERROR
        await self.vmshepherd.run(run_once=True)

        self.assertEqual(
            preset.vms,
            [VmMock(0, ['127.0.0.1'], 'test-preset', 'fedora-27', 'm1.small')]
        )


class TestVmShepherdLockingWithDummyDrivers(AsyncTestCase):

    def setUp(self):
        super().setUp()
        Drivers.flush()
        self.vmshepherd = VmShepherd(example_config)

    async def test_locking(self):
        await self.vmshepherd.preset_manager.list_presets()
        preset = self.vmshepherd.preset_manager.get_preset('test-preset')
        await asyncio.gather(
            self.vmshepherd.run(run_once=True),
            self.vmshepherd.run(run_once=True)
        )

        preset = self.vmshepherd.preset_manager.get_preset('test-preset')
        self.assertEqual(
            preset.vms, []
        )

    @expectedFailure
    async def test_example_of_bad_locking(self):
        # no lock-mechanism actually, always manage preset
        self.vmshepherd.runtime_manager.acquire_lock = Mock(return_value=futurized(True))
        self.vmshepherd.runtime_manager.release_lock = Mock(return_value=futurized(True))

        await asyncio.gather(
            self.vmshepherd.run(run_once=True),
            self.vmshepherd.worker._manage()  # explicit call of private method to bypass run_once check
        )

        preset = self.vmshepherd.preset_manager.get_preset('test-preset')
        self.assertEqual(
            preset.vms,
            [VmMock(0, ['127.0.0.1'], 'test-preset', 'fedora-27', 'm1.small')]
        )



from unittest import TestCase
from vmshepherd import VmShepherd


class TestVmShepherd(TestCase):

    def test_vmshepherd(self):
        self.assertIsNotNone(VmShepherd)
        self.assertTrue(True)



''' IaaS driver is responsible for communication with IaaS provider. To VmShepherd work,
driver should implement `create_vm`, `terminate_vms`, `list_vms` and `get_vm_info`.

Initialization - consider following config:

::

   iaas:
     driver: SomeIaaS
     auth_key: AAAA
     auth_secret: BBBB
     some_x: CCC

All params will be passed as config dict to the driver init:

'''
from .vm import Vm
from typing import Any, Dict, List


class AbstractIaasDriver:

    def __init__(self, config=None):
        pass

    async def create_vm(self, preset_name: str, image: str, flavor: str, security_groups: List=None,
                        userdata: Dict=None, key_name: str=None, availability_zone: str=None,
                        subnets: List=None) -> Any:
        """
        Create (boot) a new server.

        :arg string preset_name: Name of vm group where vm is created.
        :arg string image: Image name.
        :arg string flavor: Flavor (or instance_type in AWS) name.
        :arg list security_groups: A list of security group names.
        :arg dict userdata: A dict of arbitrary key/value metadata to store in grains.
        :arg string key_name: (optional extension) name of previously created
                      keypair to inject into the instance.
        :arg string availability_zone: Name of the availability zone for instance
                                  placement.
        :arg string subnets: List of the subnets for instance placement.

        Returns Any vm_id.
        """

        raise NotImplementedError

    async def list_vms(self, preset_name: str) -> List[Vm]:
        """
        List vms.

        :arg string preset_name: List VMs with name preset_name.

        Returns list of Vm objects.
        """

        raise NotImplementedError

    async def terminate_vm(self, vm_id: Any) -> None:
        """
        Terminates/discards vm.

        :arg string vm_id: Id Vm to terminate.
        """

        raise NotImplementedError

    async def get_vm(self, vm_id: Any) -> Vm:
        """
        Get vm info (with metadata and/or tags).

        :arg string vm_id: Id Vm to get.

        Returns list of Vm objects.
        """

        raise NotImplementedError

    def reconfigure(self, config):
        pass



import logging
from .abstract import AbstractIaasDriver
from vmshepherd.errors import IaaSError
from .vm import Vm, VmState
from vmshepherd.utils import next_id, add_async_delay


class DummyIaasUserExc(IaaSError):
    def __init__(self, message, details):
        super().__init__(message, details)


class DummyIaasVmNotFound(DummyIaasUserExc):
    def __init__(self):
        super().__init__('VMNOTFOUND', 'Vm Not Found')


class DummyIaasDriver(AbstractIaasDriver):
    """ DummyIaasDriver - mainly for test purposes.

    Implements required API of fake-in-memory-iaas (dict).
    To mimic communication with external service additional delay has been introduced.

    """

    def __init__(self, config=None):
        """
        Init: create list of vms.
        """
        self._vms = {}
        self._id_it = next_id()

    @add_async_delay
    async def create_vm(self, *, preset_name, image, flavor, security_groups=None,
                        userdata=None, key_name=None, availability_zone=None,
                        subnet=None):
        """
        Dummy create_vm func.
        """
        info = {
            'id': next(self._id_it),
            'name': preset_name,
            'ip': ['127.0.0.1'],
            'created': 0,
            'state': VmState.RUNNING,
            'flavor': flavor,
            'image': image,
            'metadata': {'test-meta': 'abctest'},
            'timed_shutdown_at': 1522753481,
            'tags': ['a-tag', 'b-tag', 'c-tag']
        }
        logging.debug('Prepare vm: %s', info)
        vm = Vm(self, **info)
        self._vms[vm.id] = vm
        logging.debug('Create: %s', vm)
        return None

    @add_async_delay
    async def list_vms(self, preset_name):
        """Dummy list_vms func"""
        return list(vm for vm in self._vms.values() if vm.name == preset_name)

    @add_async_delay
    async def terminate_vm(self, vm_id):
        """ Dummy terminate_vm func """
        if vm_id not in self._vms:
            raise DummyIaasVmNotFound()
        del self._vms[vm_id]
        return None

    @add_async_delay
    async def get_vm(self, vm_id):
        """ Dummy get_vm func """
        if vm_id not in self._vms:
            raise DummyIaasVmNotFound()
        return self._vms[vm_id]



import datetime
import logging
import time
from enum import Enum


class VmState(Enum):
    """
    Standard states for a Virtual Machine.

    :var RUNNING: Vm is running.
    :var NEARBY_SHUTDOWN: Vm is running but it's minutes are numbered.
    :var AFTER_TIME_SHUTDOWN: Vm is running but it should be dead now.
    :var TERMINATED: Vm is terminated. This node can't be started later on.
    :var PENDING: Vm is pending.
    :var UNKNOWN: Vm state is unknown.
    :var ERROR: Vm is an error state. Usually no operations can be performed
                 on the node once it ends up in the error state.
    """

    RUNNING = 'running'
    NEARBY_SHUTDOWN = 'nearby_shutdown'
    AFTER_TIME_SHUTDOWN = 'after_time_shutdown'
    TERMINATED = 'terminated'
    PENDING = 'pending'
    UNKNOWN = 'unknown'
    ERROR = 'error'


class Vm:
    """
    Virtual machine object.
    """

    def __init__(self, manager, id, name, ip, created, state=VmState.UNKNOWN, metadata=None, tags=None,
                 flavor=None, image=None, timed_shutdown_at=None):

        """ Init for VM.

        :arg AbstractIaasDriver manager: class implementing Vm actions.
        :arg string id: Vm uniq id.
        :arg string name: Vm name/preset_name ??.
        :arg list ip: Vm ip list.
        :arg datetime/int created: Timestamp when Vm was created
        :arg VmState state: Vm state.
        :arg dict metadata: Vm metadata.
        :arg list tags: Vm tags.
        :arg string flavor: Vm flavor name.
        :arg string image: Vm image name.
        :arg int timed_shutdown_at: Timestamp when VM will be terminated.
        """

        self.manager = manager
        self.id = id
        self.name = name
        self.ip = ip
        self.state = state
        self.metadata = metadata
        self.tags = tags
        self.flavor = flavor
        self.image = image
        self.timed_shutdown_at = timed_shutdown_at
        if isinstance(created, datetime.datetime):
            self.created = created
        elif isinstance(created, (int, float)):
            self.created = datetime.datetime.fromtimestamp(created)
        else:
            # actually it's not so crucial to fail
            self.created = time.time()

    def __str__(self):
        return f'VM id={self.id} preset={self.name} state={self.state} ip={self.ip}'

    def __eq__(self, other):
        for prop in ('id', 'ip', 'name', 'flavor', 'image'):
            if getattr(self, prop) != getattr(other, prop):
                return False
        return True

    def __gt__(self, other):
        return self.created > other.created

    def __hash__(self):
        return hash(self.id)

    async def terminate(self):
        """ Terminate vm.
        """
        logging.debug('Terminate: %s', self)
        self.state = VmState.TERMINATED
        return await self.manager.terminate_vm(self.id)

    def is_running(self):
        return self.state in (VmState.RUNNING, VmState.NEARBY_SHUTDOWN)

    def is_dead(self):
        return self.state in (VmState.TERMINATED, VmState.ERROR, VmState.AFTER_TIME_SHUTDOWN)

    def get_state(self):
        return self.state.value



import time
from aiounittest import futurized, AsyncTestCase
from unittest.mock import patch, Mock
from vmshepherd.http.rpc_api import RpcApi
from vmshepherd.iaas.vm import Vm, VmState


class TestRpcApi(AsyncTestCase):

    def tearDown(self):
        patch.stopall()
        super().tearDown()

    def setUp(self):
        super().setUp()
        mock_request = Mock()
        self.mock_preset_manager = Mock()
        self.mock_preset_manager.vms = [
            Vm(self, '1243454353', 'C_DEV-app-dev',
               ['10.177.51.8'], time.time(), state=VmState.RUNNING),
            Vm(self, '4535646466', 'C_DEV-app-dev',
               ['10.177.51.9'], time.time(), state=VmState.RUNNING),
            Vm(self, '5465465643', 'C_DEV-app-dev',
               ['10.177.51.10'], time.time(), state=VmState.RUNNING)
        ]
        self.mock_preset_manager.count = 3
        self.mock_preset_manager.iaas.terminate_vm.return_value = futurized('ok')
        self.mock_preset_manager.iaas.get_vm.return_value = futurized(self.mock_preset_manager.vms[0])
        mock_request.app.vmshepherd.preset_manager.get_preset.return_value = self.mock_preset_manager
        mock_request.app.vmshepherd.preset_manager.list_presets.return_value = futurized({"C_DEV-app-dev": []})
        mock_request.remote = ['10.177.51.8']
        self.rpcapi = RpcApi()
        self.rpcapi._request = mock_request
        self.mock_list_vms = {
            "1243454353": {
                "ip": "10.177.51.8",
                "state": "running"
            },
            "4535646466": {
                "ip": "10.177.51.9",
                "state": "running"
            },
            "5465465643": {
                "ip": "10.177.51.10",
                "state": "running"
            }
        }

    async def test_list_vms(self):
        self.mock_preset_manager.refresh_vms.return_value = futurized(None)
        ret = await self.rpcapi.list_vms('C_DEV-app-dev')
        self.mock_preset_manager.refresh_vms.assert_called_once_with()

        # next call should also refresh vms
        ret = await self.rpcapi.list_vms('C_DEV-app-dev')
        self.assertEqual(self.mock_preset_manager.refresh_vms.call_count, 2)

        self.assertEqual(ret[1], self.mock_list_vms)
        self.assertEqual(ret[0], 3)

    async def test_terminate_vm_success(self):
        ret = await self.rpcapi.terminate_vm('C_DEV-app-dev', 12345)
        self.assertEqual(ret, 'OK')
        self.mock_preset_manager.iaas.terminate_vm.assert_called_with(12345)

    async def test_get_vm_metadata_success(self):
        ret = await self.rpcapi.get_vm_metadata('C_DEV-app-dev', 12345)
        self.assertEqual(ret, {'tags': None, 'iaas_shutdown': None})
        self.mock_preset_manager.iaas.get_vm.assert_called_with(12345)

    async def test_list_presets_success(self):
        ret = await self.rpcapi.list_presets()
        self.assertEqual(ret, ["C_DEV-app-dev"])

    async def test_get_vm_ip(self):
        ret = await self.rpcapi.get_vm_ip('C_DEV-app-dev', '1243454353')
        self.assertEqual({'vm_ip': '10.177.51.8'}, ret)



''' Abstract class of healthcheck. There is only one method to implement,
that determines state of Vm or its underlying services.

Initialization - consider following config:

::

   healthcheck:
     driver: SomeHC
     param1: AAAA
     param2: BBBB
     some_x: CCC

All params will be passed as config dict to the driver init:

'''


class AbstractHealthcheck:

    def __init__(self, config):
        pass

    async def is_healthy(self, vm):
        ''' Checks that is a given vm is healthy

        :arg vmshepherd.iaas.Vm vm: Vm object

        Returns boolean - True means is healthy, False means unhealthy.
        '''
        raise NotImplementedError



import asyncio
import logging
from .abstract import AbstractRuntimeData


class InMemoryDriver(AbstractRuntimeData):
    ''' Simple in-memory driver for runtime data and locks managment.
    '''

    def __init__(self, instance_id, config=None):
        super().__init__(instance_id, config)
        self._storage = {}
        self._locks = {}

    async def _acquire_lock(self, name, timeout=1):
        if not self._locks.get(name):
            self._locks[name] = asyncio.Lock()
        fut = self._locks[name].acquire()
        try:
            await asyncio.wait_for(fut, timeout)
            return True
        except asyncio.TimeoutError:
            return False
        except Exception:
            logging.exception('Lock %s failed.', name)
            return False

    async def _release_lock(self, name):
        try:
            self._locks[name].release()
        except RuntimeError:
            logging.exception('Unlock %s failed.', name)

    async def _set_preset_data(self, preset_name, data):
        self._storage[preset_name] = data

    async def _get_preset_data(self, preset_name):
        return self._storage.get(preset_name, {})



import json
from pkg_resources import iter_entry_points


class Drivers:

    _loaded = {}

    @classmethod
    def get(cls, group: str, cfg: dict, **kwargs) -> object:
        _hash = hash(json.dumps(cfg, sort_keys=True))
        if _hash not in cls._loaded:
            for entry_point in iter_entry_points(group=f'vmshepherd.driver.{group}'):
                if entry_point.name == cfg['driver']:
                    _class = entry_point.load()
                    try:
                        cls._loaded[_hash] = _class(config=cfg, **kwargs)
                    except Exception as exc:
                        raise RuntimeError(f"Cannot load driver {cfg['driver']} for {group}.") from exc
                    break
            else:
                raise RuntimeError(f"Cannot find driver {cfg['driver']} for {group}.")
        return cls._loaded[_hash]

    @classmethod
    def flush(cls):
        cls._loaded = {}



''' Runtime manager or runtime data storage stores intermediate states of VmShepherd cycles
that need to be shared across multiple instances of application like:
 - healthchecks status,
 - try counts
 - time and dates of last manage

and locking mechanism.

The locking mechanism guarantees a preset is managed by one instance at a time.

Initialization - consider following config:

::

   runtime:
     driver: SomeHC
     param1: AAAA
     param2: BBBB
     some_x: CCC

All params will be passed as config dict to the driver init:
'''
import time


# TODO: it should be @dataclass
class Data:

    def __init__(self, d):
        self.last_managed = d.get('last_managed', {}).get('time', 0)
        self.last_managed_by = d.get('last_managed', {}).get('id', '')
        self.failed_checks = d.get('failed_checks', {})
        self.iaas = d.get('iaas', {})

    def dump(self):
        return {'iaas': self.iaas, 'failed_checks': self.failed_checks}

    def __str__(self):
        return str(self.__dict__)


class AbstractRuntimeData:

    def __init__(self, instance_id, config=None):
        self.instance_id = instance_id
        self.preset_lock_ttl = config.get('preset_lock_ttl', 60) if config else 60

    async def _acquire_lock(self, name: str) -> bool:
        ''' Locks preset during manage process, any other VMgr instance won't disrubt it.

        :arg string name: Lock name in particular preset name.

        If lock was acquired it should return True, otherwise False.
        '''
        raise NotImplementedError

    async def _release_lock(self, name: str) -> None:
        ''' Unlocks presets, other VMgr instance will be able to manage it.

        :arg string name: Lock name in particular preset name.
        :arg string preset_name: Preset name
        '''
        raise NotImplementedError

    async def _set_preset_data(self, preset_name: str, data: dict) -> None:
        ''' Saves preset's runtime data - fail counters, times

        :arg string preset_name: Preset name
        :arg dict data: Set of counters and other volatile data
        '''
        raise NotImplementedError

    async def _get_preset_data(self, preset_name: str) -> dict:
        ''' Gets preset's runtime data eg. fail counters, times

        :arg string preset_name: Preset name
        '''
        raise NotImplementedError

    async def get_preset_data(self, name):
        response = await self._get_preset_data(name)
        return Data(response)

    async def set_preset_data(self, name, data):
        commit = {
            'last_managed': {
                'time': time.time(),
                'id': self.instance_id
            }
        }
        commit.update(data.dump())
        return await self._set_preset_data(name, commit)

    async def acquire_lock(self, name):
        return await self._acquire_lock(name)

    async def release_lock(self, name):
        return await self._release_lock(name)

    def reconfigure(self, config):
        pass



import asyncio
import errno
import logging
import os
import tempfile
from .abstract import AbstractConfigurationDriver
from asyncio.subprocess import PIPE
from vmshepherd.utils import async_load_from_yaml_file


class GitRepoDriver(AbstractConfigurationDriver):

    def __init__(self, config, runtime, defaults):
        super().__init__(runtime, defaults)
        self._clone_dir = config.get('clone_dir', tempfile.TemporaryDirectory(prefix='vmshepherd').name)
        self._repos = config['repositories']
        self._specs = {}
        self.lock = asyncio.Lock()

    async def _get_preset_spec(self, preset_name: str):
        return self._specs[preset_name]

    async def _list(self):
        await self.reload()
        return self._specs.keys()

    async def reload(self):
        _tmp_specs = {}
        self._assure_clone_dir_exists()
        for name, repo in self._repos.items():
            try:
                path = os.path.join(self._clone_dir, name)
                await self._clone_or_update(path, repo)
                added = await self._load_repo(name, path)
                _tmp_specs.update(added)
            except Exception:
                logging.exception('GitReposDriver: Could not load %s from %s', name, repo)
        self._specs = _tmp_specs

    async def _load_repo(self, repo_name, path):
        loaded = {}
        for item in os.scandir(path):
            if os.path.isfile(item.path) and os.path.splitext(item.path)[1] == '.conf':
                config = await async_load_from_yaml_file(item.path)
                # prepend repo name to preset_name
                config['name'] = preset_name = f"{repo_name}.{config['name']}"
                config['userdata_source_root'] = path
                loaded[preset_name] = config
        return loaded

    async def _clone_or_update(self, path, repo):
        try:
            await self.lock.acquire()
            if os.path.exists(path):
                cmd = ['git', '-C', path, 'pull']
            else:
                cmd = ['git', 'clone', repo, path]
            process = await asyncio.create_subprocess_exec(
                *cmd, stdout=PIPE, stderr=PIPE
            )
            stdout, stderr = await process.communicate()
            logging.debug('GitRepo code=%s stdout: %s stderr: %s', process.returncode, stdout, stderr)
            if process.returncode != 0:
                logging.error('Git error: %s %s %s', path, repo, stderr)
                raise RuntimeError(f'Could not fetch presets ({path}) from {repo}')
        finally:
            self.lock.release()

    def _assure_clone_dir_exists(self):
        try:
            os.makedirs(self._clone_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    def reconfigure(self, config, defaults):
        super().reconfigure(config, defaults)
        self._clone_dir = config.get('clone_dir', self._clone_dir)
        self._repos = config.get('repositories', self._repos)



import io
import re
from os.path import join, dirname, abspath
from setuptools import setup, find_packages


def read(name):
    here = abspath(dirname(__file__))
    return io.open(
        join(here, name), encoding='utf8'
    ).read()


setup(
    name="vmshepherd",
    version="1.3.3",
    author='Dreamlab - PaaS KRK',
    author_email='paas-support@dreamlab.pl',
    url='https://github.com/Dreamlab/vmshepherd',
    description='Cluster manager',
    long_description='%s\n%s' % (
        read('README.rst'),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=read('requirements.txt').split('\n'),
    zip_safe=False,
    entry_points={
        'console_scripts': ['vmshepherd = vmshepherd.__main__:main'],
        'vmshepherd.driver.iaas': [
            'DummyIaasDriver = vmshepherd.iaas:DummyIaasDriver',
            'OpenStackDriver = vmshepherd.iaas:OpenStackDriver'
        ],
        'vmshepherd.driver.presets': [
            'DirectoryDriver = vmshepherd.presets:DirectoryDriver',
            'GitRepoDriver = vmshepherd.presets:GitRepoDriver'
        ],
        'vmshepherd.driver.runtime': ['InMemoryDriver = vmshepherd.runtime:InMemoryDriver'],
        'vmshepherd.driver.healthcheck': ['HttpHealthcheck = vmshepherd.healthcheck:HttpHealthcheck']
    },
    keywords=[
        'cluster', 'preset',
        'iaas', 'cloud',
        'openstack', 'nova',
        'aws', 'amazon', 'ec2'
    ],
    dependency_links=['https://pypi.python.org/pypi'],
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: POSIX'
    ]
)



from vmshepherd.app import VmShepherd  # noqa
from vmshepherd.iaas import AbstractIaasDriver  # noqa
from vmshepherd.presets import AbstractConfigurationDriver  # noqa
from vmshepherd.runtime import AbstractRuntimeData  # noqa



import copy
from aiohttp_jsonrpc import handler
from functools import wraps


class RpcApi(handler.JSONRPCView):
    """
    RPC Api to manage virtual machines
    """

    def __init__(self, allowed_methods=None):
        self.METHOD_PREFIX = ''
        self.allowed_methods = allowed_methods

    async def handler(self, request):
        self._request = request
        return await self

    def enabled_checker(func):
        """ Access decorator which checks if a RPC method is enabled by our configuration
        """

        @wraps(func)
        def wrap(self, *args, **kwargs):
            if self.allowed_methods and isinstance(self.allowed_methods, list) and func.__name__ not in self.allowed_methods:
                raise Exception("Method {} is disabled".format(func.__name__))
            return func(self, *args, **kwargs)

        return wrap

    @enabled_checker
    async def list_presets(self):
        """
        Listing presets

        :return:  (list of presets)

        :rtype: list

        Sample response:
            ``["preset1", "preset2"]``
        """
        presets = await self.request.app.vmshepherd.preset_manager.list_presets()
        return list(presets.keys())

    @enabled_checker
    async def list_vms(self, preset):
        """ Listing virtual machines in a given preset

        :arg string preset: preset name
        :return:  (Size of a preset, list of virtual machines)

            - first element of a tuple is a size of virtual machines in a preset
            - second element is a dict which contains all Virtual Machines, where every element of this dict looks like that:

              ``{ "VIRTUAL_MACHINE_ID": { "ip": "IP_ADDR", "state": "VM_STATE" }``

        :rtype: tuple

        Sample response:
            ``( 1, {'180aa486-ee46-4628-ab1c-f4554b63231': {'ip': '172.1.1.2', 'state': 'running'}} )``
        """
        vmshepherd = self.request.app.vmshepherd
        preset = vmshepherd.preset_manager.get_preset(preset)
        await preset.refresh_vms()
        result_vms = {vm.id: {'ip': vm.ip[0], 'state': vm.state.value} for vm in preset.vms}
        return preset.count, result_vms

    @enabled_checker
    async def terminate_vm(self, preset, vm_id):
        """ Discard vm in specified preset

        :arg string preset: preset name
        :arg int vm_id: Virtual Machine id
        :return: 'OK'

        Sample response:
           ``OK``
        """
        vmshepherd = self.request.app.vmshepherd
        preset = vmshepherd.preset_manager.get_preset(preset)
        await preset.iaas.terminate_vm(vm_id)
        return 'OK'

    @enabled_checker
    async def get_vm_metadata(self, preset, vm_id):
        """ Get vm metadata

        :arg string preset: preset name
        :arg int vm_id: Virtual Machine id
        :return:  Metadata for Virtual Machine
        :rtype: dict

        Sample response:
           ``{ 'time_shutdown' : "12312312321' }``
        """
        vmshepherd = self.request.app.vmshepherd
        preset = vmshepherd.preset_manager.get_preset(preset)
        vm_info = await preset.iaas.get_vm(vm_id)
        ret_info = copy.deepcopy(vm_info.metadata) if vm_info.metadata else {}
        ret_info['tags'] = vm_info.tags
        ret_info['iaas_shutdown'] = vm_info.timed_shutdown_at
        return ret_info

    @enabled_checker
    async def get_vm_ip(self, preset_name, vm_id):
        """ Get vm ip

        :arg string preset: preset name
        :arg int vm_id: Virtual Machine id
        :return:  Vm Ip
        :rtype: string
        """
        vmshepherd = self.request.app.vmshepherd
        preset = vmshepherd.preset_manager.get_preset(preset_name)
        # check in cache
        for vm in preset.vms:
            if vm_id == vm.id:
                return {'vm_ip': vm.ip[0]}
        # retrieve real time data
        vm_info = await preset.iaas.get_vm(vm_id)
        return {
            'vm_ip': vm_info.ip[0]
        }



import asyncio
import logging
import time


class Worker:
    """ Preset management class """

    ERROR = 'ERROR'
    OK = 'OK'

    def __init__(self, runtime, presets, interval=1, autostart=True):
        self.presets = presets
        self.runtime = runtime
        self._interval = interval
        self._running = False
        self._start_time = 0
        if autostart:
            asyncio.ensure_future(self.run_forever())

    async def run_forever(self):
        self._forever = True
        while self._forever:
            await self.run_once()
            await asyncio.sleep(self._interval)

    def stop(self):
        self._forever = False

    async def run_once(self):
        if self._running:
            logging.info('Already running since %s', self._start_time)
            return
        self._running = True

        self._start_time = time.time()
        logging.info('VmShepherd start cycle...')

        result, cnt_managed, cnt_presets, cnt_failed_presets = self.ERROR, 0, -1, -1
        try:
            result, cnt_presets, cnt_managed, cnt_failed_presets = await self._manage()
        except Exception:
            logging.exception('Error while running')
        finally:
            logging.info(
                'VmShepherd end cycle result=%s presets=%s managed=%s failed_presets=%s time=%.2f',
                result, cnt_presets, cnt_managed, cnt_failed_presets, time.time() - self._start_time
            )
            self._running = False
        return result

    async def _manage(self):

        presets = await self.presets.list_presets(refresh=True)
        cnt_presets, cnt_managed, cnt_failed_presets = len(presets), 0, 0

        for name, preset in presets.items():

            try:
                async with preset as locked:
                    if locked:
                        cnt_managed += 1
                        await preset.manage()
            except Exception:
                logging.exception('Error managing %s', name)
                cnt_failed_presets += 1

        return self.OK, cnt_presets, cnt_managed, cnt_failed_presets



import aiohttp
from aiounittest import AsyncTestCase, futurized
from collections import namedtuple
from unittest.mock import Mock, patch
from vmshepherd.healthcheck import HttpHealthcheck

VmMock = namedtuple('VmMock', 'id ip name image flavor')
HttpResp = namedtuple('HttpResp', 'status')


class TestHttpHealthcheck(AsyncTestCase):

    def setUp(self):
        super().setUp()
        self.healthcheck = HttpHealthcheck()

    async def test_fail_without_ip(self):
        vm_without_ip = VmMock(0, [], 'test-preset', 'fedora-27', 'm1.small')
        self.assertFalse(await self.healthcheck.is_healthy(vm_without_ip))

    async def test_check_ok(self):
        vm = VmMock(0, ['127.0.0.1'], 'test-preset', 'fedora-27', 'm1.small')
        resp_200 = HttpResp(200)

        with patch('aiohttp.ClientSession.request', Mock()) as mock_request:
            mock_request.return_value = futurized(resp_200)
            self.assertTrue(await self.healthcheck.is_healthy(vm))
            mock_request.assert_called_once_with(method='GET', url='http://127.0.0.1:80/')

    async def test_check_fail(self):
        vm = VmMock(0, ['127.0.0.1'], 'test-preset', 'fedora-27', 'm1.small')

        with patch('aiohttp.ClientSession.request', Mock()) as mock_request:
            mock_request.return_value = futurized(aiohttp.client_exceptions.ClientError())
            self.assertFalse(await self.healthcheck.is_healthy(vm))
            mock_request.assert_called_once_with(method='GET', url='http://127.0.0.1:80/')



import argparse
import asyncio
import signal
from functools import partial
from vmshepherd.app import VmShepherd
from vmshepherd.utils import load_config_file


def main():
    args = get_args()
    config = load_config_file(args.config_file)
    config['log_level'] = args.log_level
    config['autostart'] = False
    config['http']['listen_port'] = int(args.port)
    vmshepherd = VmShepherd(config)

    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGHUP, partial(hup_handler, vmshepherd, args.config_file))

    task = asyncio.ensure_future(vmshepherd.run(args.run_once))
    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt as e:
        print("VmShepherd shutting down...")
        task.cancel()
    finally:
        loop.close()


def hup_handler(app, config_file, *args, **kwargs):
    config = load_config_file(config_file)
    app.reload(config)


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--config-file', type=str, help='config file', required=True)
    parser.add_argument('-l', '--log-level', type=str, help='logging level', default='info')
    parser.add_argument('-p', '--port', type=int, help='port for panel and api', default=8888)
    parser.add_argument('-1', '--run-once', dest='run_once', action='store_true')
    args = parser.parse_args()
    return args


if __name__ == '__main__':
    main()
